import asyncio
from typing import AsyncIterator, Union
from rath.links.base import AsyncTerminatingLink
from rath.operation import GraphQLResult, Operation
from graphql import FieldNode, GraphQLSchema, OperationType, validate


def target_from_node(node: FieldNode) -> str:
    return (
        node.alias.value if hasattr(node, "alias") and node.alias else node.name.value
    )


class AsyncMockResolver:
    def __getitem__(self, key):
        return getattr(self, f"resolve_{key}")

    def __contains__(self, key):
        return hasattr(self, f"resolve_{key}")


class AsyncMockLink(AsyncTerminatingLink):
    def __init__(
        self,
        resolver: Union[AsyncMockResolver, dict] = {},
        query_resolver: Union[AsyncMockResolver, dict] = None,
        mutation_resolver: Union[AsyncMockResolver, dict] = None,
        subscription_resolver: Union[AsyncMockResolver, dict] = None,
    ) -> None:
        self.query_resolver = query_resolver or resolver
        self.mutation_resolver = mutation_resolver or resolver
        self.subscription_resolver = subscription_resolver or resolver

    async def aquery(self, operation: Operation) -> GraphQLResult:

        if operation.node.operation == OperationType.QUERY:
            futures = []

            for op in operation.node.selection_set.selections:
                assert (
                    op.name.value in self.query_resolver
                ), f"Mocked Resolver for Query '{op.name.value}' not in resolver: {self.query_resolver}  for AsyncMockLink"
                futures.append(self.query_resolver[op.name.value](operation))

            resolved = await asyncio.gather(*futures)
            return GraphQLResult(
                data={
                    target_from_node(op): resolved[i]
                    for i, op in enumerate(operation.node.selection_set.selections)
                }
            )

        if operation.node.operation == OperationType.MUTATION:
            futures = []

            for op in operation.node.selection_set.selections:
                assert (
                    op.name.value in self.mutation_resolver
                ), f"Mocked Resolver for Mutation {op.name.value} not in resolver: {self.mutation_resolver} for AsyncMockLink"
                futures.append(self.mutation_resolver[op.name.value](operation))

            resolved = await asyncio.gather(*futures)
            return GraphQLResult(
                data={
                    target_from_node(op): resolved[i]
                    for i, op in enumerate(operation.node.selection_set.selections)
                }
            )

    async def asubscribe(self, operation: Operation) -> AsyncIterator[GraphQLResult]:
        if operation.node.operation == OperationType.SUBSCRIPTION:
            futures = []
            assert (
                len(operation.node.selection_set.selections) == 1
            ), "Only one Subscription at a time possible"

            op = operation.node.selection_set.selections[0]
            assert (
                op.name.value in self.subscription_resolver
            ), f"Mocked Resolver for Subscription {op.name.value} not in resolver: {self.subscription_resolver} for AsyncMockLink"

            async for event in self.subscription_resolver[op.name.value](operation):
                yield GraphQLResult(data={target_from_node(op): event})

        else:
            raise NotImplementedError("Only subscription are mocked")

    def query(self, operation: Operation) -> GraphQLResult:
        raise NotImplementedError("This is a async Terminating Link")

    def subscribe(self, operation: Operation) -> GraphQLResult:
        raise NotImplementedError("This is a async Terminating Link")



import asyncio
from rath.links.base import TerminatingLink
from typing import (
    AsyncIterator,
    Dict,
    Any,
    Iterator,
    Optional,
    List,
    Union,
    Callable,
    Awaitable,
)
from rath.operation import GraphQLResult, opify
from contextvars import ContextVar


class Rath:
    def __init__(
        self,
        link: TerminatingLink,
        register=False,
        autoconnect=True,
    ) -> None:
        """Initialize a Rath client

        Rath takes a instance of TerminatingLink and creates an interface around it
        to enable easy usage of the GraphQL API.

        Args:
            link (TerminatingLink): A terminating link or a composed link.
            register (bool, optional): Register as a global rath (knowing the risks). Defaults to False.
            autoconnect (bool, optional): [description]. Defaults to True.
        """
        self.link = link
        self.autoconnect = autoconnect
        self.link(self)

        if register:
            set_current_rath(self)

    async def aexecute(
        self,
        query: str,
        variables: Dict[str, Any] = None,
        headers: Dict[str, Any] = {},
        operation_name=None,
        timeout=None,
        **kwargs,
    ) -> GraphQLResult:

        op = opify(query, variables, headers, operation_name, **kwargs)

        if timeout:
            return await asyncio.wait_for(self.link.aquery(op), timeout)

        return await self.link.aquery(
            op,
        )

    def execute(
        self,
        query: str,
        variables: Dict[str, Any] = None,
        headers: Dict[str, Any] = {},
        operation_name=None,
        **kwargs,
    ) -> GraphQLResult:
        op = opify(query, variables, headers, operation_name, **kwargs)

        return self.link.query(
            op,
        )

    def subscribe(
        self,
        query: str,
        variables: Dict[str, Any] = None,
        headers: Dict[str, Any] = {},
        operation_name=None,
        **kwargs,
    ) -> Iterator[GraphQLResult]:

        op = opify(query, variables, headers, operation_name, **kwargs)
        for res in self.link.subscribe(op):
            yield res

    async def asubscribe(
        self,
        query: str,
        variables: Dict[str, Any] = None,
        headers: Dict[str, Any] = {},
        operation_name=None,
        **kwargs,
    ) -> AsyncIterator[GraphQLResult]:

        op = opify(query, variables, headers, operation_name, **kwargs)
        async for res in self.link.asubscribe(op):
            yield res

    async def __aenter__(self):
        await self.link.__aenter__()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.link.__aexit__(exc_type, exc_val, exc_tb)

    def __enter__(self):
        self.link.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.link.__exit__(exc_type, exc_val, exc_tb)

    def connect(self):
        return self.__enter__()

    def disconnect(self):
        return self.__exit__(None, None, None)


CURRENT_RATH = None


def get_current_rath(**kwargs):
    global CURRENT_RATH
    assert CURRENT_RATH is not None, "No current rath set"
    return CURRENT_RATH


def set_current_rath(rath):
    global CURRENT_RATH
    CURRENT_RATH = rath



from pydantic import BaseModel
from herre.herre import Herre, get_current_herre
from rath.links.auth import AuthTokenLink


class HerreAuthTokenLink(AuthTokenLink):
    def __init__(self, *args, herre: Herre = None, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.herre = herre or get_current_herre()
        self.token_loader = self.herre.aget_token



from dataclasses import dataclass, field
from glob import glob
from typing import Optional, Dict, Any
from graphql.language import OperationDefinitionNode, parse
from graphql import (
    DocumentNode,
    GraphQLSchema,
    get_introspection_query,
    get_operation_ast,
    parse,
    build_ast_schema,
)


@dataclass
class Context:
    headers: Optional[Dict[str, str]] = field(default_factory=dict)
    files: Optional[Dict[str, Any]] = field(default_factory=dict)
    kwargs: Optional[Dict[str, Any]] = field(default_factory=dict)


@dataclass
class Extensions:
    pollInterval: Optional[int] = None
    maxPolls: Optional[int] = None


@dataclass
class Operation:
    document_node: DocumentNode
    node: OperationDefinitionNode
    document: str
    variables: Dict[str, Any]
    operation_name: str
    extensions: Extensions
    context: Context


@dataclass
class GraphQLResult:
    data: Dict[str, Any]


class GraphQLException(Exception):
    pass


def opify(
    query: str,
    variables: Dict[str, Any] = None,
    headers: Dict[str, Any] = {},
    operation_name: Optional[str] = None,
    **kwargs,
) -> Operation:

    document = parse(query)
    op = get_operation_ast(document, operation_name)
    assert op, f"No operation named {operation_name}"
    return Operation(
        node=op,
        document=query,
        document_node=document,
        variables=variables or {},
        operation_name=operation_name,
        extensions={},
        context=Context(headers=headers, kwargs=kwargs),
    )



import asyncio
from concurrent.futures import ThreadPoolExecutor
from rath.links.base import ContinuationLink
from rath.operation import GraphQLResult, Operation
from koil import unkoil, unkoil_gen, Koil


class SwitchAsyncLink(ContinuationLink):
    """Swithc Async Link

    Async switches allow to use async resolvers in an sync context.
    It uses the koil library to facilitate the switch.

    Koil can use unstateful links (no "async with" logic), by simply calling
    asyncio.run(). Or spin up a thread with the async logic for stateful links
    (with "async with" logic). It also takes care of stoppin the thread when
    the link is disconnected.

    This allows us to use the same logic for both sync and async links, where
    a connection is required.

    Check the koil documentation to learn more about how koil does it in detail.

    Args:
        ContinuationLink (_type_): _description_
    """

    def __init__(self, **koilparams):
        super().__init__()
        self.__koil = Koil(**koilparams)

    async def aquery(self, operation: Operation) -> GraphQLResult:
        return await self.next.aquery(operation)

    async def asubscribe(self, operation: Operation) -> GraphQLResult:
        async for result in self.next.asubscribe(operation):
            yield result

    def query(self, operation: Operation) -> GraphQLResult:
        return unkoil(self.next.aquery, operation)

    def subscribe(self, operation: Operation) -> GraphQLResult:
        for result in unkoil_gen(self.next.asubscribe, operation):
            yield result

    def __enter__(self) -> None:
        self.__koil.__enter__()  # spin up a new thread if needed (if we are already in a koiled environment, use that one)
        unkoil(self.next.__aenter__)

    def __exit__(self, *args, **kwargs) -> None:
        unkoil(self.next.__aexit__, *args, **kwargs)
        self.__koil.__exit__(None, None, None)  # cleanup thread if we had to spin it up


class SwitchSyncLink(ContinuationLink):
    def __init__(self, excecutor=None) -> None:
        self.excecutor = excecutor or ThreadPoolExecutor()
        self._lock = asyncio.Lock()
        self.connected = False
        super().__init__()

    async def __aenter__(self) -> None:
        self.e = self.excecutor.__enter__()
        self.next.__enter__()
        self.connected = True

    async def aquery(self, operation: Operation) -> GraphQLResult:
        async with self._lock:
            if not self.connected:
                await self.aconnect()

        return await asyncio.wrap_future(self.e.submit(self.next.query, operation))

    async def asubscribe(self, operation: Operation) -> GraphQLResult:
        raise NotImplementedError(
            "We need to fiqure this out yet. Normally a __next__ call here would be enough"
        )

    def query(self, operation: Operation) -> GraphQLResult:
        return self.next.query(operation)

    def subscribe(self, operation: Operation) -> GraphQLResult:
        for result in self.next.subscribe(operation):
            yield result

    async def __aexit__(self, *args, **kwargs) -> None:
        self.e = self.excecutor.__exit__()
        self.next.__exit__(*args, **kwargs)
        self.connected = False



from pydantic.main import BaseModel, ModelMetaclass
from turms.types.operation import OperationMeta
from rath.rath import Rath, get_current_rath


class GraphQLOperation(BaseModel, metaclass=OperationMeta):
    class Meta:
        abstract = True


class GraphQLQuery(GraphQLOperation):
    class Meta:
        abstract = True


class GraphQLMutation(GraphQLOperation):
    class Meta:
        abstract = True


class GraphQLSubscription(GraphQLOperation):
    class Meta:
        abstract = True



def clean_dict(obj, func):
    """
    This method scrolls the entire 'obj' to delete every key for which the 'callable' returns
    True

    :param obj: a dictionary or a list of dictionaries to clean
    :param func: a callable that takes a key in argument and return True for each key to delete
    """
    if isinstance(obj, dict):
        # the call to `list` is useless for py2 but makes
        # the code py2/py3 compatible

        for key, value in dict(**obj).items():
            if func(key, value):
                del obj[key]
            else:
                clean_dict(obj[key], func)
    elif isinstance(obj, list):
        for i in reversed(range(len(obj))):
            clean_dict(obj[i], func)

    else:
        # neither a dict nor a list, do nothing
        pass



from typing import AsyncIterator
from graphql import (
    GraphQLSchema,
    build_ast_schema,
    build_schema,
    get_introspection_query,
    validate,
)
from graphql.language.parser import parse
from rath.links.base import ContinuationLink
from rath.links.errors import ContinuationLinkError
from rath.operation import GraphQLResult, Operation, opify
from .parsing import ParsingLink
from glob import glob


def schemify(
    schema_dsl: str = None, schema_glob: str = None, schema_url: str = None
) -> GraphQLSchema:
    if schema_dsl:
        return build_ast_schema(parse(schema_dsl))
    if schema_glob:
        schema_glob = glob(schema_glob, recursive=True)
        dsl_string = ""
        for file in schema_glob:
            with open(file, "r") as f:
                if file.endswith(".graphql"):
                    dsl_string += f.read()

        assert dsl_string, f"No schema found in glob {schema_glob}"
        return build_ast_schema(parse(dsl_string))

    raise NotImplementedError("Please provide either a dsl or glob")


class ValidationError(ContinuationLinkError):
    pass


class ValidatingLink(ContinuationLink):
    def __init__(
        self,
        schema_dsl: str = None,
        schema_glob: str = None,
        allow_introspection=True,
    ) -> None:
        if schema_dsl or schema_glob:
            self.schema = schemify(schema_dsl, schema_glob)

        self.allow_introspection = allow_introspection

    async def aload_schema(self, operation: Operation) -> None:
        if self.allow_introspection:
            introspect_operation = opify(get_introspection_query())
            introspect_operation.context = operation.context
            introspect_operation.extensions = operation.extensions

            schema_result = await self.next.aquery(introspect_operation)
            self.schema = build_schema(schema_result.data)

    def load_schema(self, operation: Operation) -> None:
        if self.allow_introspection:
            introspect_operation = opify(get_introspection_query())
            introspect_operation.context = operation.context
            introspect_operation.extensions = operation.extensions

            schema_result = self.next.query(introspect_operation)
            self.schema = build_schema(schema_result.data)

    def validate(self, operation: Operation):
        errors = validate(self.schema, operation.document_node)

        if len(errors) > 0:
            raise ValidationError(
                f"{operation} does not comply with the schema!\n Errors: \n\n"
                + "\n".join([e.message for e in errors])
            )

    async def aquery(self, operation: Operation) -> GraphQLResult:
        if not self.schema:
            await self.aload_schema(operation)

        self.validate(operation)
        return await self.next.aquery(operation)

    async def asubscribe(self, operation: Operation) -> AsyncIterator[GraphQLResult]:
        if not self.schema:
            await self.aload_schema(operation)

        self.validate(operation)
        async for result in self.next.asubscribe(operation):
            yield result

    def query(self, operation: Operation) -> GraphQLResult:
        if not self.schema:
            self.load_schema(operation)

        self.validate(operation)
        return self.next.query(operation)

    def subscribe(self, operation: Operation) -> AsyncIterator[GraphQLResult]:
        if not self.schema:
            self.load_schema(operation)

        self.validate(operation)
        for result in self.next.subscribe(operation):
            yield result



from rath.operation import Operation
from rath.links.parsing import ParsingLink


class DictingLink(ParsingLink):
    def parse(self, operation: Operation) -> Operation:
        shrinked_variables = {
            key: var.dict()
            for key, var in operation.variables.items()
            if hasattr(var, "dict")
        }
        operation.variables.update(shrinked_variables)
        return operation

    async def aparse(self, operation: Operation) -> Operation:
        shrinked_variables = {
            key: var.dict()
            for key, var in operation.variables.items()
            if hasattr(var, "dict")
        }
        operation.variables.update(shrinked_variables)
        return operation



import asyncio
from os import link
from typing import List
from rath.links.base import ContinuationLink, Link, TerminatingLink
from rath.operation import Operation


class ComposedLink(TerminatingLink):
    def __init__(self, links: List[Link]):
        assert isinstance(
            links[-1], TerminatingLink
        ), "Last link must be a TerminatingLink"
        for link in links[:-1]:
            assert isinstance(
                link, ContinuationLink
            ), f"All but the last must be ContinuationLinks: check {link}"
        self.links = links
        self.first_link = links[0]

    def __call__(self, rath):
        for i in range(len(self.links) - 1):
            self.links[i](rath, self.links[i + 1])

        self.links[-1](rath)  # last one gets only rath

    async def __aenter__(self):
        await self.first_link.__aenter__()

    async def __aexit__(self, *args, **kwargs):
        await self.first_link.__aexit__(*args, **kwargs)

    def __enter__(self):
        self.first_link.__enter__()

    def __exit__(self, *args, **kwargs):
        self.first_link.__exit__(*args, **kwargs)

    async def aquery(self, operation: Operation):
        return await self.first_link.aquery(operation)

    async def asubscribe(self, operation: Operation):
        async for result in self.first_link.asubscribe(operation):
            yield result

    def query(self, operation: Operation):
        return self.first_link.query(operation)

    def subscribe(self, operation: Operation):
        for result in self.first_link.subscribe(operation):
            yield result


def compose(*links: List[Link]) -> ComposedLink:
    """
    Composes a list of Links into a single Link.
    """

    return ComposedLink(links)



import asyncio
from pydantic import BaseModel
from fakts.fakts import Fakts, get_current_fakts
from rath.links.aiohttp import AIOHttpLink
from rath.operation import Operation


class FaktsAioHttpLinkConfig(BaseModel):
    endpoint_url: str


class FaktsAioHttpLink(AIOHttpLink):
    def __init__(self, *args, fakts: Fakts = None, fakts_key="rath", **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._lock = None
        self.fakts = fakts or get_current_fakts()
        self.fakts_key = fakts_key
        self.config = None

    def configure(self, fakts: FaktsAioHttpLinkConfig):
        self.url = fakts.endpoint_url
        self.config = fakts

    async def aquery(self, operation: Operation, **kwargs) -> None:
        if not self._lock:
            self._lock = asyncio.Lock()

        async with self._lock:
            if not self.config:
                fakts = await self.fakts.aget(self.fakts_key)
                self.configure(FaktsAioHttpLinkConfig(**fakts))

        return await super().aquery(operation, **kwargs)



import asyncio
from typing import AsyncIterator, Union
from rath.links.base import AsyncTerminatingLink
from rath.operation import GraphQLResult, Operation
from graphql import FieldNode, GraphQLSchema, OperationType, validate
import uuid
from rath.links.errors import LinkNotConnectedError


def target_from_node(node: FieldNode) -> str:
    return (
        node.alias.value if hasattr(node, "alias") and node.alias else node.name.value
    )


class AsyncMockResolver:
    def __getitem__(self, key):
        return getattr(self, f"resolve_{key}")

    def __contains__(self, key):
        return hasattr(self, f"resolve_{key}")


class AsyncStatefulMockLink(AsyncTerminatingLink):
    def __init__(
        self,
        resolver: Union[AsyncMockResolver, dict] = {},
        query_resolver: Union[AsyncMockResolver, dict] = None,
        mutation_resolver: Union[AsyncMockResolver, dict] = None,
        subscription_resolver: Union[AsyncMockResolver, dict] = None,
        schema: GraphQLSchema = None,
    ) -> None:
        self.query_resolver = query_resolver or resolver
        self.mutation_resolver = mutation_resolver or resolver
        self.subscription_resolver = subscription_resolver or resolver
        self.schema = schema
        self.connected = False
        self.futures = {}

    async def __aenter__(self) -> None:
        self.connected = True
        self._inqueue = asyncio.Queue()
        self.connection_task = asyncio.create_task(self.resolving())

    async def __aexit__(self, *args, **kwargs) -> None:
        self.connected = False
        self.connection_task.cancel()

        try:
            await self.connection_task
        except asyncio.CancelledError:
            pass

    async def resolving(self):
        while True:
            operation, id = await self._inqueue.get()

            resolve_futures = []

            try:
                if operation.node.operation == OperationType.QUERY:

                    for op in operation.node.selection_set.selections:
                        assert (
                            op.name.value in self.query_resolver
                        ), f"Mocked Resolver for Query '{op.name.value}' not in resolver: {self.query_resolver}  for AsyncMockLink"

                        resolve_futures.append(
                            self.query_resolver[op.name.value](operation)
                        )

                if operation.node.operation == OperationType.MUTATION:

                    for op in operation.node.selection_set.selections:
                        assert (
                            op.name.value in self.mutation_resolver
                        ), f"Mocked Resolver for Mutation {op.name.value} not in resolver: {self.mutation_resolver} for AsyncMockLink"
                        resolve_futures.append(
                            self.mutation_resolver[op.name.value](operation)
                        )

                resolved = await asyncio.gather(*resolve_futures)
                self.futures[id].set_result(
                    GraphQLResult(
                        data={
                            target_from_node(op): resolved[i]
                            for i, op in enumerate(
                                operation.node.selection_set.selections
                            )
                        }
                    )
                )

            except Exception as e:
                self.futures[id].set_exception(e)

            self._inqueue.task_done()

    async def submit(self, o, id):
        await self._inqueue.put((o, id))

    async def aquery(self, operation: Operation) -> GraphQLResult:
        if self.connected == False:
            raise LinkNotConnectedError("AsyncMockLink is not connected")

        uniqueid = uuid.uuid4()
        self.futures[uniqueid] = asyncio.Future()
        await self.submit(operation, uniqueid)
        return await self.futures[uniqueid]

    async def asubscribe(self, operation: Operation) -> AsyncIterator[GraphQLResult]:
        if operation.node.operation == OperationType.SUBSCRIPTION:
            futures = []
            assert (
                len(operation.node.selection_set.selections) == 1
            ), "Only one Subscription at a time possible"

            op = operation.node.selection_set.selections[0]
            assert (
                op.name.value in self.subscription_resolver
            ), f"Mocked Resolver for Subscription {op.name.value} not in resolver: {self.subscription_resolver} for AsyncMockLink"

            async for event in self.subscription_resolver[op.name.value](operation):
                yield GraphQLResult(data={target_from_node(op): event})

        else:
            raise NotImplementedError("Only subscription are mocked")

    def query(self, operation: Operation) -> GraphQLResult:
        raise NotImplementedError("This is a async Terminating Link")

    def subscribe(self, operation: Operation) -> GraphQLResult:
        raise NotImplementedError("This is a async Terminating Link")



from typing import Callable, List
from koil.loop import koil, koil_gen
from rath.operation import Operation
from rath.links.base import TerminatingLink


class SplitLink(TerminatingLink):
    def __init__(
        self,
        left: TerminatingLink,
        right: TerminatingLink,
        split: Callable[[Operation], bool],
    ) -> None:
        super().__init__()
        assert isinstance(left, TerminatingLink), "left must be a TerminatingLink"
        assert isinstance(right, TerminatingLink), "right must be a TerminatingLink"
        self.left = left
        self.right = right
        self.split = split

    def __call__(self, rath):
        self.left(rath)
        self.right(rath)

    async def aquery(self, operation: Operation) -> Operation:
        future = (
            self.left.aquery(operation)
            if self.split(operation)
            else self.right.aquery(operation)
        )
        return await future

    async def asubscribe(self, operation: Operation) -> Operation:
        iterator = (
            self.left.asubscribe(operation)
            if self.split(operation)
            else self.right.asubscribe(operation)
        )

        async for res in iterator:
            yield res

    def query(self, operation: Operation) -> Operation:
        future = (
            self.left.aquery(operation)
            if self.split(operation)
            else self.right.aquery(operation)
        )

        return koil(future)

    def subscribe(self, operation: Operation) -> Operation:
        iterator = (
            self.left.asubscribe(operation)
            if self.split(operation)
            else self.right.asubscribe(operation)
        )

        for res in koil_gen(iterator):
            yield res

    async def __aenter__(self):
        await self.left.__aenter__()
        await self.right.__aenter__()

    async def __aexit__(self, *args, **kwargs):
        await self.left.__aexit__(*args, **kwargs)
        await self.right.__aexit__(*args, **kwargs)

    def __exit__(self, *args, **kwargs):
        self.left.__exit__(*args, **kwargs)
        self.right.__axit__(*args, **kwargs)

    def __enter__(self, *args, **kwargs):
        self.left.__enter__(*args, **kwargs)
        self.right.__enter__(*args, **kwargs)


def split(
    left: TerminatingLink, right: TerminatingLink, split: Callable[[Operation], bool]
):
    """
    Splits a Link into two paths. Acording to a predicate function. If predicate returns
    true, the operation is sent to the left path, otherwise to the right path.
    """
    return SplitLink(left, right, split)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['rath',
 'rath.fakts',
 'rath.herre',
 'rath.links',
 'rath.links.testing',
 'rath.turms']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.8.1,<4.0.0',
 'graphql-core>=3.2.0,<4.0.0',
 'koil==0.1.77',
 'pydantic>=1.9.0,<2.0.0']

setup_kwargs = {
    'name': 'rath',
    'version': '0.1.18',
    'description': 'aiohttp powered apollo like graphql client',
    'long_description': '# rath\n\n[![codecov](https://codecov.io/gh/jhnnsrs/rath/branch/master/graph/badge.svg?token=UGXEA2THBV)](https://codecov.io/gh/jhnnsrs/rath)\n[![PyPI version](https://badge.fury.io/py/rath.svg)](https://pypi.org/project/rath/)\n[![Maintenance](https://img.shields.io/badge/Maintained%3F-yes-green.svg)](https://pypi.org/project/rath/)\n![Maintainer](https://img.shields.io/badge/maintainer-jhnnsrs-blue)\n[![PyPI pyversions](https://img.shields.io/pypi/pyversions/rath.svg)](https://pypi.python.org/pypi/rath/)\n[![PyPI status](https://img.shields.io/pypi/status/rath.svg)](https://pypi.python.org/pypi/rath/)\n[![PyPI download month](https://img.shields.io/pypi/dm/rath.svg)](https://pypi.python.org/pypi/rath/)\n\n### DEVELOPMENT\n\n## Inspiration\n\nRath is like Apollo, but for python. It adheres to the design principle of Links and enables complex GraphQL\nsetups, like seperation of query and subscription endpoints, dynamic token loading, etc..\n\n## Installation\n\n```bash\npip install rath\n```\n\n## Usage Example\n\n```python\nfrom rath.links.auth import AuthTokenLink\nfrom rath.links.aiohttp import AioHttpLink\nfrom rath.links import compose, split\nfrom rath.gql import gql\n\nasync def aload_token():\n    return "SERVER_TOKEN"\n\n\nauth = AuthTokenLink(token_loader=aload_token)\nlink = AioHttpLink(url="https://api.spacex.land/graphql/")\n\n\nrath = Rath(links=compose(auth,link))\nrath.connect()\n\n\nquery = """query TestQuery {\n  capsules {\n    id\n    missions {\n      flight\n    }\n  }\n}\n"""\n\nresult = rath.execute(query)\n```\n\nThis example composes both the AuthToken and AioHttp link: During each query the Bearer headers are set to the retrieved token, on authentication fail (for example if Token Expired) the\nAuthToken automatically refetches the token and retries the query.\n\n## Async Usage\n\nRath is build with koil, for async/sync compatibility but also exposed a complete asynhronous api, also it is completely threadsafe\n\n```python\nfrom rath.links.auth import AuthTokenLink\nfrom rath.links.aiohttp import AioHttpLink\nfrom rath.links import compose, split\nfrom rath.gql import gql\n\nasync def aload_token():\n    return "SERVER_TOKEN"\n\n\nauth = AuthTokenLink(token_loader=aload_token)\nlink = AioHttpLink(url="https://api.spacex.land/graphql/")\n\n\nasync def main():\n  rath = Rath(links=compose(auth,link))\n  await rath.aconnect()\n\n\n  query = """query TestQuery {\n    capsules {\n      id\n      missions {\n        flight\n      }\n    }\n  }\n  """\n\n  result = await rath.aexecute(query)\n\nasyncio.run(main())\n```\n\n## Usage of Async Links in Sync Environments\n\nLinks can either have a synchronous or asynchronous interface (inheriting from SyncLink or AsyncLink). Using an Async Link from a Sync\ncontext however is not possible without switching context. For this purpose exist SwitchLinks that can either switch from sync to async\nor back.\n\n```python\n\nupload_files = UploadFilesSyncLink(bucket="lala")\nswitch = SwitchAsyncLink()\nlink = AioHttpLink(url="https://api.spacex.land/graphql/")\n\nrath = Rath(link=compose(upload_files, switch, link))\n\n```\n\n## Example Transport Switch\n\n```python\nlink = SplitLink(\n  AioHttpLink(url="https://api.spacex.land/graphql/"),\n  WebsocketLink(url="ws://api.spacex.land/graphql/",\n  lamda o: o.node.operation == OperationType.SUBSCRIPTION\n)\n\nrath = Rath(link=link)\n\n```\n\n## Included Links\n\n- Validating Link (validate query against local schema (or introspect the schema))\n- Reconnecting WebsocketLink\n- AioHttpLink (supports multipart uploads)\n- SplitLink (allows to split the terminating link - Subscription into WebsocketLink, Query, Mutation into Aiohttp)\n- AuthTokenLink (Token insertion with automatic refresh)\n\n## Dynamic Configuration\n\nrath follows some design principles of fakts for asynchronous configuration retrieval, and provides some examplary links\n\n## Authentication\n\nIf you want to use rath with herre for getting access_tokens in oauth2/openid-connect scenarios, there is also a herre link\nin this repository\n\n### Why Rath\n\nWell "apollo" is already taken as a name, and rath (according to wikipedia) is an etruscan deity identified with Apollo.\n\n## Rath + Turms\n\nRath works especially well with turms generated typed operations:\n\n```python\nimport asyncio\nfrom examples.api.schema import aget_capsules\nfrom rath.rath import Rath\nfrom rath.links.aiohttp import AIOHttpLink\nfrom rath.links.auth import AuthTokenLink\nfrom rath.links.compose import compose\n\n\nasync def token_loader():\n    return ""\n\n\nlink = compose(\n    AuthTokenLink(token_loader), AIOHttpLink("https://api.spacex.land/graphql/")\n)\n\n\nRATH = Rath(\n    link=link,\n    register=True, # allows global access (singleton-antipattern, but rath has no state)\n)\n\n\nasync def main():\n    capsules = await aget_capsules() # fully typed pydantic powered dataclasses generated through turms\n    print(capsules)\n\n\nasyncio.run(main())\n\n```\n\n## Examples\n\nThis github repository also contains an example client with a turms generated query with the public SpaceX api, as well as a sample of the generated api.\n',
    'author': 'jhnnsrs',
    'author_email': 'jhnnsrs@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import asyncio
from http import HTTPStatus
import json
from typing import Any, AsyncIterator, Dict

import aiohttp
from graphql import OperationType
from rath.operation import GraphQLException, GraphQLResult, Operation
from rath.links.base import AsyncTerminatingLink
from rath.links.errors import AuthenticationError


class AIOHttpLink(AsyncTerminatingLink):

    auth_errors = [HTTPStatus.FORBIDDEN]

    def __init__(self, url: str = "") -> None:
        self.url = url

    async def aquery(self, operation: Operation) -> GraphQLResult:
        payload = {"query": operation.document}

        if len(operation.context.files.items()) > 0:
            payload["variables"] = operation.variables

            files = operation.context.files
            data = aiohttp.FormData()

            file_map = {str(i): [path] for i, path in enumerate(files)}

            # Enumerate the file streams
            # Will generate something like {'0': <_io.BufferedReader ...>}
            file_streams = {str(i): files[path] for i, path in enumerate(files)}
            operations_str = json.dumps(payload)

            data.add_field(
                "operations", operations_str, content_type="application/json"
            )
            file_map_str = json.dumps(file_map)
            data.add_field("map", file_map_str, content_type="application/json")

            for k, v in file_streams.items():
                data.add_field(
                    k,
                    v,
                    filename=getattr(v, "name", k),
                )

            post_kwargs: Dict[str, Any] = {"data": data}

        else:
            payload["variables"] = operation.variables
            post_kwargs = {"json": payload}

        async with aiohttp.ClientSession(headers=operation.context.headers) as session:
            try:
                async with session.post(self.url, **post_kwargs) as response:

                    if response.status == HTTPStatus.OK:
                        result = await response.json()

                    if response.status in self.auth_errors:
                        raise AuthenticationError(
                            f"Token Expired Error {operation.context.headers}"
                        )

                    json_response = await response.json()

                    if "errors" in json_response:
                        raise GraphQLException(
                            "\n".join([e["message"] for e in json_response["errors"]])
                        )

                    if "data" not in json_response:

                        raise Exception(
                            f"Response does not contain data {json_response}"
                        )

                    return GraphQLResult(data=json_response["data"])

            except aiohttp.client_exceptions.InvalidURL as e:
                raise Exception(f"Invalid URL {self.url}") from e

    async def asubscribe(self, operation: Operation) -> AsyncIterator[GraphQLResult]:
        if operation.node.operation == OperationType.SUBSCRIPTION:
            raise NotImplementedError(
                "Aiohttp Transport does not support subscriptions"
            )

        if operation.node.operation == OperationType.QUERY:
            if operation.extensions.pollInterval:
                while True:
                    yield await self.aquery(operation)
                    await asyncio.sleep(operation.extensions.pollInterval)
            else:
                raise NotImplementedError(
                    "If you didn't specify a pollInterval you cannot use subscribe to this query"
                )

    def query(self, operation: Operation) -> GraphQLResult:
        raise NotImplementedError()

    def subscribe(self, operation: Operation) -> GraphQLResult:
        raise NotImplementedError()
